@@ -1,7 +1,11 @@
 import os
 import numpy as np
 import pandas as pd
-from crypto_strategy.data import download_crypto_history
+from crypto_strategy.data import (
+    save_stats,
+    get_acc_returns,
+    download_crypto_history
+)
 from .base import (
     breakout_strategy,
     vol_filter, ang_filter,
@@ -98,7 +102,7 @@ def create_bo_variables(**kwargs):
 
 def create_bo_variables_ts_stop(**kwargs):
     if kwargs:
-        assert kwargs.get('long_window') and kwargs.get('short_window') and kwargs.get('ts_stop'), \
+        assert kwargs.get('long_window') and kwargs.get('short_window') and kwargs.get('ts_stop'),\
             'BO strategy doesn\'t have config params provided'
         variables = dict(
             long_window=kwargs['long_window'],
@@ -127,10 +131,12 @@ class BestBoStrategy(BestStrategy):
     def __init__(self, symbols: list, freq: str, res_dir: str,
                  flag_filter: str = None,
                  flag_ts_stop: bool = False,
+                 flag_acc_return: bool = True,
                  strategy: str = 'bo',
                  ):
         super().__init__(symbols, freq, res_dir, flag_filter, strategy)
         self.flag_ts_stop = flag_ts_stop
+        self.flag_acc_return = flag_acc_return
         self.generate_best_params()
 
     def _get_strategy(self, strategy):
@@ -156,12 +162,12 @@ class BestBoStrategy(BestStrategy):
         total_best_params = pd.DataFrame(total_best_params)
         total_best_params['sharpe'] = (total_best_params['sharpe'] + 0.05).round(1)
         total_best_params['gap'] = total_best_params['long_window'] - total_best_params['short_window']
-        if 'advstex_ts_stop' in total_best_params.columns:
+        if 'ohlcstx_sl_stop' in total_best_params.columns:
             total_best_params = (
                 total_best_params
                 .query('gap > 0')
                 .sort_values(
-                    by=['sharpe', 'gap', 'short_window', 'advstex_ts_stop'],
+                    by=['sharpe', 'gap', 'short_window', 'ohlcstx_sl_stop'],
                     ascending=[True, True, False, False])
             )
         else:
@@ -173,7 +179,6 @@ class BestBoStrategy(BestStrategy):
                     ascending=[True, True, False]
                     )
             )
-
         print(total_best_params)
         return total_best_params.tail(1).to_dict(orient='records')[0] if not total_best_params.empty else None
 
@@ -207,7 +212,7 @@ class BestBoStrategy(BestStrategy):
             variables = self._get_variables(
                 long_window=best_params['long_window'],
                 short_window=best_params['short_window'],
-                ts_stop=best_params['advstex_ts_stop']
+                ts_stop=best_params['ohlcstx_sl_stop']
             )
         else:
             variables = self._get_variables(
@@ -219,18 +224,23 @@ class BestBoStrategy(BestStrategy):
             multiplier=best_params.get('vol_multiplier'),
             threshold=best_params.get('ang_threshold')
             )
-        portfolio = self.strategy.backtest(self.ohlcv, freq=self.freq, variables=variables, filters=filters)
         filename = f'''{symbol}-{self.freq}-{self.strategy_name}-{best_params['long_window']}-{best_params['short_window']}-'''
         if self.flag_ts_stop:
-            filename += f'''ts-{best_params['advstex_ts_stop']:.2f}-'''
+            filename += f'''ts-{best_params['ohlcstx_sl_stop']:.2f}-'''
         if self.flag_filter == 'vol':
             filename += f'''{self.flag_filter}-{best_params['vol_timeperiod']}-{best_params['vol_multiplier']}-{self.date_str}.pkl'''
         elif self.flag_filter == 'ang':
             filename += f'''{self.flag_filter}-{best_params['ang_timeperiod']}-{best_params['ang_threshold']}-{self.date_str}.pkl'''
         else:
             filename += f'''{self.date_str}.pkl'''
+        portfolio = self.strategy.backtest(self.ohlcv, freq=self.freq, variables=variables, filters=filters)
+        stats = portfolio.stats()
+        if self.flag_acc_return:
+            acc_returns = get_acc_returns(portfolio.daily_returns())
+            stats = stats.append(pd.Series(acc_returns))
         filename = os.path.join(self.output_path, filename)
-        portfolio.stats().to_pickle(filename)
+        save_stats(stats, self.output_path, filename)
+        print(f'The stats are saved to {self.output_path}/{filename}')
 
     def generate_best_params(self):
         print(f'The search for {self.symbols} starts now')
